MODULE traldf_iso_tile
   !!======================================================================
   !!                   ***  MODULE  traldf_iso  ***
   !! Ocean  tracers:  horizontal component of the lateral tracer mixing trend
   !!======================================================================
   !! History :  OPA  ! 1994-08  (G. Madec, M. Imbard)
   !!            8.0  ! 1997-05  (G. Madec)  split into traldf and trazdf
   !!            NEMO ! 2002-08  (G. Madec)  Free form, F90
   !!            1.0  ! 2005-11  (G. Madec)  merge traldf and trazdf :-)
   !!            3.3  ! 2010-09  (C. Ethe, G. Madec) Merge TRA-TRC
   !!            3.7  ! 2014-01  (G. Madec, S. Masson)  restructuration/simplification of aht/aeiv specification
   !!             -   ! 2014-02  (F. Lemarie, G. Madec)  triad operator (Griffies) + Method of Stabilizing Correction
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   tra_ldf_iso_tile   : update the tracer trend with the horizontal component of a iso-neutral laplacian operator
   !!                   and with the vertical part of the isopycnal or geopotential s-coord. operator 
   !!----------------------------------------------------------------------
   USE par_kind
   USE len_oce            ! ocean dynamics and active tracers
   USE tiletype
   !
   USE in_out_manager ! I/O manager

   IMPLICIT NONE

   PUBLIC   tra_ldf_iso_tile   ! routine called by step.F90

   !! * Substitutions
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OCE 4.0 , NEMO Consortium (2018)
   !! $Id$
   !! Software governed by the CeCILL licence     (./LICENSE)
   !!----------------------------------------------------------------------
CONTAINS

  SUBROUTINE tra_ldf_iso_tile( tile, ln_traldf_msc, ln_traldf_blp, ln_traldf_lap, ln_zps, ln_isfcav, kjpt, kpass, p2dt, &
      &                   e1t, e1u, e1v, e2t, e2u, e2v, e1e2t, e1_e2v, e2_e1u, r1_e1e2t, mbku, mbkv, miku, mikv, &
      &                   umask, vmask, wmask, e3t_n, e3u_n, e3v_n, e3w_n, wslpi, wslpj, uslp, vslp, &
      &                   pahu, pahv, pgu, pgv, pgui, pgvi, ptb , ptbb, pta, &
      &                   l_hst, l_ptr, pakz, pah_wslp2, pftu, pftv )

      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_ldf_iso_tile  ***
      !!
      !! ** Purpose :   Compute the before horizontal tracer (t & s) diffusive 
      !!      trend for a laplacian tensor (ezxcept the dz[ dz[.] ] term) and 
      !!      add it to the general trend of tracer equation.
      !!
      !! ** Method  :   The horizontal component of the lateral diffusive trends 
      !!      is provided by a 2nd order operator rotated along neural or geopo-
      !!      tential surfaces to which an eddy induced advection can be added
      !!      It is computed using before fields (forward in time) and isopyc-
      !!      nal or geopotential slopes computed in routine ldfslp.
      !!
      !!      1st part :  masked horizontal derivative of T  ( di[ t ] )
      !!      ========    with partial cell update if ln_zps=T
      !!                  with top     cell update if ln_isfcav
      !!
      !!      2nd part :  horizontal fluxes of the lateral mixing operator
      !!      ========    
      !!         zftu =  pahu e2u*e3u/e1u di[ tb ]
      !!               - pahu e2u*uslp    dk[ mi(mk(tb)) ]
      !!         zftv =  pahv e1v*e3v/e2v dj[ tb ]
      !!               - pahv e2u*vslp    dk[ mj(mk(tb)) ]
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1e2t*e3t) {  di-1[ zftu ] +  dj-1[ zftv ]  }
      !!      Add this trend to the general trend (ta,sa):
      !!         ta = ta + difft
      !!
      !!      3rd part: vertical trends of the lateral mixing operator
      !!      ========  (excluding the vertical flux proportional to dk[t] )
      !!      vertical fluxes associated with the rotated lateral mixing:
      !!         zftw = - {  mi(mk(pahu)) * e2t*wslpi di[ mi(mk(tb)) ]
      !!                   + mj(mk(pahv)) * e1t*wslpj dj[ mj(mk(tb)) ]  }
      !!      take the horizontal divergence of the fluxes:
      !!         difft = 1/(e1e2t*e3t) dk[ zftw ]
      !!      Add this trend to the general trend (ta,sa):
      !!         pta = pta + difft
      !!
      !! ** Action :   Update pta arrays with the before rotated diffusion
      !!----------------------------------------------------------------------
      TYPE(tile_type)                      , INTENT(in   ) ::   tile
      LOGICAL                              , INTENT(in   ) ::   ln_traldf_msc, ln_traldf_blp, ln_traldf_lap, ln_zps, ln_isfcav
      INTEGER                              , INTENT(in   ) ::   kjpt       ! number of tracers passed in
      INTEGER                              , INTENT(in   ) ::   kpass      ! =1/2 first or second passage
      REAL(wp)                             , INTENT(in   ) ::   p2dt       ! 2 . dt 
      REAL(wp), DIMENSION(jpi,jpj)         , INTENT(in   ) ::   e1t, e1u, e1v, e2t, e2u, e2v, e1e2t, e1_e2v, e2_e1u, r1_e1e2t     
      INTEGER, DIMENSION(jpi,jpj)          , INTENT(in   ) ::   mbku, mbkv, miku, mikv
      REAL(wp), DIMENSION(jpi,jpj,jpk)     , INTENT(in   ) ::   umask, vmask, wmask
      REAL(wp), DIMENSION(jpi,jpj,jpk)     , INTENT(in   ) ::   e3t_n, e3u_n, e3v_n, e3w_n, wslpi, wslpj, uslp, vslp

      REAL(wp), DIMENSION(jpi,jpj,jpk)     , INTENT(in   ) ::   pahu, pahv ! eddy diffusivity at u- and v-points  [m2/s]
      REAL(wp), DIMENSION(jpi,jpj    ,kjpt), INTENT(in   ) ::   pgu, pgv   ! tracer gradient at pstep levels
      REAL(wp), DIMENSION(jpi,jpj,    kjpt), INTENT(in   ) ::   pgui, pgvi ! tracer gradient at top   levels
      REAL(wp), DIMENSION(jpi,jpj,jpk,kjpt), INTENT(in   ) ::   ptb        ! tracer (kpass=1) or laplacian of tracer (kpass=2)
      REAL(wp), DIMENSION(jpi,jpj,jpk,kjpt), INTENT(in   ) ::   ptbb       ! tracer (only used in kpass=2)
      REAL(wp), DIMENSION(jpi,jpj,jpk,kjpt), INTENT(inout) ::   pta        ! tracer trend

      LOGICAL                              , INTENT(in   ) ::   l_ptr      ! flag to compute poleward transport
      LOGICAL                              , INTENT(in   ) ::   l_hst      ! flag to compute heat transport
      REAL(wp), DIMENSION(jpi,jpj,jpk)     , INTENT(inout) ::   pakz, pah_wslp2 !        
      REAL(wp), DIMENSION(jpi,jpj,jpk,kjpt), INTENT(inout) ::   pftu, pftv ! diagnostic outputs

      !
      INTEGER  ::  ji, jj, jk, jn   ! dummy loop indices
      INTEGER  ::  ikt
      INTEGER  ::  ierr             ! local integer
      REAL(wp) ::  zmsku, zahu_w, zabe1, zcof1, zcoef3   ! local scalars
      REAL(wp) ::  zmskv, zahv_w, zabe2, zcof2, zcoef4   !   -      -
      REAL(wp) ::  zcoef0, ze3w_2, zsign, z2dt, z1_2dt   !   -      -

      REAL(wp) ,ALLOCATABLE, DIMENSION(:,:)     ::   zdkt, zdk1t, z2d      ! Make these use global scratch at some point
      REAL(wp) ,ALLOCATABLE, DIMENSION(:,:,:)   ::   zdit, zdjt, zftu, zftv, ztfw,  zakz, zah_wslp2 

      ALLOCATE(zdkt (tile%jsih:tile%jeih, tile%jsjh:tile%jejh))
      ALLOCATE(zdk1t(tile%jsih:tile%jeih, tile%jsjh:tile%jejh))
      ALLOCATE(z2d  (tile%jsih:tile%jeih, tile%jsjh:tile%jejh))

      ALLOCATE(zdit(     tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))
      ALLOCATE(zdjt(     tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))
      ALLOCATE(zftu(     tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))
      ALLOCATE(zftv(     tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))
      ALLOCATE(ztfw(     tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))
      ALLOCATE(zakz(     tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))
      ALLOCATE(zah_wslp2(tile%jsih:tile%jeih, tile%jsjh:tile%jejh,tile%jskh:tile%jekh))

write(0,*)"tile%jsih",tile%jsih
write(0,*)"tile%jeih",tile%jeih
write(0,*)"tile%jsjh",tile%jsjh
write(0,*)"tile%jejh",tile%jejh
write(0,*)"tile%jskh",tile%jskh
write(0,*)"tile%jekh",tile%jekh

!$ACC KERNELS
      !!----------------------------------------------------------------------
      !
      !   
      !
      !                                            ! set time step size (Euler/Leapfrog)
      z1_2dt = 1._wp / p2dt   !  note this was dependent on kt and neuler before
      !
      IF( kpass == 1 ) THEN   ;   zsign =  1._wp      ! bilaplacian operator require a minus sign (eddy diffusivity >0)
      ELSE                    ;   zsign = -1._wp
      ENDIF
         
      !!----------------------------------------------------------------------
      !!   0 - calculate  ah_wslp2 and akz
      !!----------------------------------------------------------------------
      !
      IF( kpass == 1 ) THEN                  !==  first pass only  ==!
         !
         DO jk = tile % jsk_2, tile % jekp1 
            DO jj = tile % jsj, tile % jej
               DO ji = tile % jsi, tile % jei   ! vector opt.

                  !
                  zmsku = wmask(ji,jj,jk) / MAX(   umask(ji  ,jj,jk-1) + umask(ji-1,jj,jk)          &
                     &                           + umask(ji-1,jj,jk-1) + umask(ji  ,jj,jk) , 1._wp  )
                  zmskv = wmask(ji,jj,jk) / MAX(   vmask(ji,jj  ,jk-1) + vmask(ji,jj-1,jk)          &
                     &                           + vmask(ji,jj-1,jk-1) + vmask(ji,jj  ,jk) , 1._wp  )
                     !
                  zahu_w = (   pahu(ji  ,jj,jk-1) + pahu(ji-1,jj,jk)    &
                     &       + pahu(ji-1,jj,jk-1) + pahu(ji  ,jj,jk)  ) * zmsku
                  zahv_w = (   pahv(ji,jj  ,jk-1) + pahv(ji,jj-1,jk)    &
                     &       + pahv(ji,jj-1,jk-1) + pahv(ji,jj  ,jk)  ) * zmskv
                     !
                  zah_wslp2(ji,jj,jk) = zahu_w * wslpi(ji,jj,jk) * wslpi(ji,jj,jk)   &
                     &               + zahv_w * wslpj(ji,jj,jk) * wslpj(ji,jj,jk)
               END DO
            END DO
         END DO
         !
         IF( ln_traldf_msc ) THEN                ! stabilizing vertical diffusivity coefficient
            DO jk = tile % jsk_2, tile % jekp1 
               DO jj = tile % jsj, tile % jej
                  DO ji = tile % jsi, tile % jei
                     zakz(ji,jj,jk) = 0.25_wp * (                                                                     &
                        &              ( pahu(ji  ,jj,jk) + pahu(ji  ,jj,jk-1) ) / ( e1u(ji  ,jj) * e1u(ji  ,jj) )   &
                        &            + ( pahu(ji-1,jj,jk) + pahu(ji-1,jj,jk-1) ) / ( e1u(ji-1,jj) * e1u(ji-1,jj) )   &
                        &            + ( pahv(ji,jj  ,jk) + pahv(ji,jj  ,jk-1) ) / ( e2v(ji,jj  ) * e2v(ji,jj  ) )   &
                        &            + ( pahv(ji,jj-1,jk) + pahv(ji,jj-1,jk-1) ) / ( e2v(ji,jj-1) * e2v(ji,jj-1) )   )
                  END DO
               END DO
            END DO
            !
            IF( ln_traldf_blp ) THEN                ! bilaplacian operator
               DO jk = tile % jsk_2, tile % jekp1 
                  DO jj = tile % jsj, tile % jej
                     DO ji = tile % jsi, tile % jei
                        zakz(ji,jj,jk) = 16._wp * zah_wslp2(ji,jj,jk)   &
                           &          * (  zakz(ji,jj,jk) + zah_wslp2(ji,jj,jk) / ( e3w_n(ji,jj,jk) * e3w_n(ji,jj,jk) )  )
                     END DO
                  END DO
               END DO
            ELSEIF( ln_traldf_lap ) THEN              ! laplacian operator
               DO jk = tile % jsk_2, tile % jekp1
                  DO jj = tile % jsj, tile % jej
                     DO ji = tile % jsi, tile % jei
                        ze3w_2 = e3w_n(ji,jj,jk) * e3w_n(ji,jj,jk)
                        zcoef0 = p2dt * (  zakz(ji,jj,jk) + zah_wslp2(ji,jj,jk) / ze3w_2  )
                        zakz(ji,jj,jk) = MAX( zcoef0 - 0.5_wp , 0._wp ) * ze3w_2 * z1_2dt
                     END DO
                  END DO
               END DO
           ENDIF
           !
         ELSE                                    ! 33 flux set to zero with zakz=zah_wslp2 ==>> computed in full implicit
           DO jk = tile % jsk_2, tile % jekp1
              DO jj = tile % jsj, tile % jej
                 DO ji = tile % jsi, tile % jei
                    zakz(ji,jj,jk) = zah_wslp2(ji,jj,jk)
                 END DO
               END DO
            END DO
         ENDIF
      ENDIF

! Save zakz and zah_wslp2 (private arrays) in non-overlapping elements of akz and ah_wslp2 arrays (for re-use in trazdf) 
      DO jk = tile % jsk, tile % jek
         DO jj = tile % jsj, tile % jej
            DO ji = tile % jsi, tile % jei
               pakz(ji,jj,jk) = zakz(ji,jj,jk) ; pah_wslp2(ji,jj,jk) = zah_wslp2(ji,jj,jk)
            END DO
         END DO
      END DO

      !
      !                                                          ! ===========
      DO jn = 1, kjpt                                            ! tracer loop
         !                                                       ! ===========
         !                                               
         !!----------------------------------------------------------------------
         !!   I - masked horizontal derivative 
         !!----------------------------------------------------------------------

         ! Horizontal tracer gradient 
         DO jk = tile % jskm1, tile % jekp1
            DO jj = tile % jsjm1, tile % jej
               DO ji = tile % jsim1, tile % jei   ! vector opt.
                  zdit(ji,jj,jk) = ( ptb(ji+1,jj  ,jk,jn) - ptb(ji,jj,jk,jn) ) * umask(ji,jj,jk)
                  zdjt(ji,jj,jk) = ( ptb(ji  ,jj+1,jk,jn) - ptb(ji,jj,jk,jn) ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
         IF( ln_zps ) THEN      ! botton and surface ocean correction of the horizontal gradient
            DO jj = tile % jsjm1, tile % jej              ! bottom correction (partial bottom cell)
               DO ji = tile % jsim1, tile % jei   ! vector opt.
                  IF( tile % jskm1 <= mbku(ji,jj) .AND. mbku(ji,jj) <= tile % jekp1 ) zdit(ji,jj,mbku(ji,jj)) = pgu(ji,jj,jn)          
                  IF( tile % jskm1 <= mbkv(ji,jj) .AND. mbkv(ji,jj) <= tile % jekp1 ) zdjt(ji,jj,mbkv(ji,jj)) = pgv(ji,jj,jn)
               END DO
            END DO
            IF( ln_isfcav ) THEN      ! first wet level beneath a cavity
               DO jj = tile % jsjm1, tile % jej
                  DO ji = tile % jsim1, tile % jei   ! vector opt.
                     IF( tile % jskm1 <= miku(ji,jj) .AND. miku(ji,jj) <= tile % jekp1 ) zdit(ji,jj,miku(ji,jj)) = pgui(ji,jj,jn)          
                     IF( tile % jskm1 <= mikv(ji,jj) .AND. mikv(ji,jj) <= tile % jekp1 ) zdjt(ji,jj,mikv(ji,jj)) = pgvi(ji,jj,jn)
                  END DO
               END DO
            ENDIF
         ENDIF
         !
         !!----------------------------------------------------------------------
         !!   II - horizontal trend  (full)
         !!----------------------------------------------------------------------
         !
         DO jk = tile % jsk, tile % jek                                 ! Horizontal slab
            !
            !                             !== Vertical tracer gradient (loops expanded for low-level tiling)
            DO jj = tile % jsjm1, tile % jejp1
              DO ji = tile % jsim1, tile % jeip1
                zdk1t(ji,jj) = ( ptb(ji,jj,jk,jn) - ptb(ji,jj,jk+1,jn) ) * wmask(ji,jj,jk+1)     ! level jk+1
              END DO 
            END DO
            !
            IF( jk == 1 ) THEN   
              DO jj = tile % jsjm1, tile % jejp1
                DO ji = tile % jsim1, tile % jeip1
                  zdkt(ji,jj) = zdk1t(ji,jj)    ! surface: zdkt(jk=1)=zdkt(jk=2)
                END DO 
              END DO
            ELSE                 
              DO jj = tile % jsjm1, tile % jejp1
                DO ji = tile % jsim1, tile % jeip1
                  zdkt(ji,jj) = ( ptb(ji,jj,jk-1,jn) - ptb(ji,jj,jk,jn) ) * wmask(ji,jj,jk)
                END DO 
              END DO
            ENDIF
            DO jj = tile % jsjm1, tile % jej            !==  Horizontal fluxes
               DO ji = tile % jsim1, tile % jei   ! vector opt.
                  zabe1 = pahu(ji,jj,jk) * e2_e1u(ji,jj) * e3u_n(ji,jj,jk)
                  zabe2 = pahv(ji,jj,jk) * e1_e2v(ji,jj) * e3v_n(ji,jj,jk)
                  !
                  zmsku = 1._wp / MAX(  wmask(ji+1,jj,jk  ) + wmask(ji,jj,jk+1)   &
                     &             + wmask(ji+1,jj,jk+1) + wmask(ji,jj,jk  ), 1._wp )
                  !
                  zmskv = 1._wp / MAX(  wmask(ji,jj+1,jk  ) + wmask(ji,jj,jk+1)   &
                     &             + wmask(ji,jj+1,jk+1) + wmask(ji,jj,jk  ), 1._wp )
                  !
                  zcof1 = - pahu(ji,jj,jk) * e2u(ji,jj) * uslp(ji,jj,jk) * zmsku
                  zcof2 = - pahv(ji,jj,jk) * e1v(ji,jj) * vslp(ji,jj,jk) * zmskv
                  !
                  zftu(ji,jj,jk) = (  zabe1 * zdit(ji,jj,jk)   &
                     &               + zcof1 * (  zdkt (ji+1,jj) + zdk1t(ji,jj)      &
                     &                          + zdk1t(ji+1,jj) + zdkt (ji,jj)  )  ) * umask(ji,jj,jk)
                  zftv(ji,jj,jk) = (  zabe2 * zdjt(ji,jj,jk)   &
                     &               + zcof2 * (  zdkt (ji,jj+1) + zdk1t(ji,jj)      &
                     &                          + zdk1t(ji,jj+1) + zdkt (ji,jj)  )  ) * vmask(ji,jj,jk)                  
               END DO
            END DO
            !
            DO jj = tile % jsj , tile % jej          !== horizontal divergence and add to pta
               DO ji = tile % jsi, tile % jei    ! vector opt.
                  pta(ji,jj,jk,jn) = pta(ji,jj,jk,jn) + zsign * (  zftu(ji,jj,jk) - zftu(ji-1,jj,jk)      &
                     &                                           + zftv(ji,jj,jk) - zftv(ji,jj-1,jk)  )   &
                     &                                        * r1_e1e2t(ji,jj) / e3t_n(ji,jj,jk)
               END DO
            END DO
         END DO                                        !   End of slab  

         !!----------------------------------------------------------------------
         !!   III - vertical trend (full)
         !!----------------------------------------------------------------------
         !
         ! Vertical fluxes
         ! ---------------
         !                          ! Surface and bottom vertical fluxes set to zero
         IF ( tile % jsk == 1  ) THEN
            DO jj = tile % jsj, tile % jej
               DO ji = tile % jsi, tile % jei   ! vector opt.      
                 ztfw(ji,jj,1) = 0._wp   
               END DO 
             END DO 
         END IF    
         IF ( tile % jek == jpkm1 ) THEN
            DO jj = tile % jsj, tile % jej
               DO ji = tile % jsi, tile % jei   ! vector opt.      
                 ztfw(ji,jj,jpk) = 0._wp   
               END DO 
             END DO 
         END IF    

         DO jk = tile % jsk_2, tile % jekp1           ! interior (2=<jk=<jpk-1)
            DO jj = tile % jsj, tile % jej
               DO ji = tile % jsi, tile % jei   ! vector opt.
                  !
                  zmsku = wmask(ji,jj,jk) / MAX(   umask(ji  ,jj,jk-1) + umask(ji-1,jj,jk)          &
                     &                           + umask(ji-1,jj,jk-1) + umask(ji  ,jj,jk) , 1._wp  )
                  zmskv = wmask(ji,jj,jk) / MAX(   vmask(ji,jj  ,jk-1) + vmask(ji,jj-1,jk)          &
                     &                           + vmask(ji,jj-1,jk-1) + vmask(ji,jj  ,jk) , 1._wp  )
                     !
                  zahu_w = (   pahu(ji  ,jj,jk-1) + pahu(ji-1,jj,jk)    &
                     &       + pahu(ji-1,jj,jk-1) + pahu(ji  ,jj,jk)  ) * zmsku
                  zahv_w = (   pahv(ji,jj  ,jk-1) + pahv(ji,jj-1,jk)    &
                     &       + pahv(ji,jj-1,jk-1) + pahv(ji,jj  ,jk)  ) * zmskv
                     !
                  zcoef3 = - zahu_w * e2t(ji,jj) * zmsku * wslpi (ji,jj,jk)   !wslpi & j are already w-masked
                  zcoef4 = - zahv_w * e1t(ji,jj) * zmskv * wslpj (ji,jj,jk)
                  !
                  ztfw(ji,jj,jk) = zcoef3 * (   zdit(ji  ,jj  ,jk-1) + zdit(ji-1,jj  ,jk)      &
                     &                        + zdit(ji-1,jj  ,jk-1) + zdit(ji  ,jj  ,jk)  )   &
                     &           + zcoef4 * (   zdjt(ji  ,jj  ,jk-1) + zdjt(ji  ,jj-1,jk)      &
                     &                        + zdjt(ji  ,jj-1,jk-1) + zdjt(ji  ,jj  ,jk)  )
               END DO
            END DO
         END DO

         !                                !==  add the vertical 33 flux  ==!
         IF( ln_traldf_lap ) THEN               ! laplacian case: eddy coef = zah_wslp2 - zakz
            DO jk = tile % jsk_2, tile % jekp1       
               DO jj = tile % jsj, tile % jej
                  DO ji = tile % jsi, tile % jei
                     ztfw(ji,jj,jk) = ztfw(ji,jj,jk) + e1e2t(ji,jj) / e3w_n(ji,jj,jk) * wmask(ji,jj,jk)   &
                        &                            * ( zah_wslp2(ji,jj,jk) - zakz(ji,jj,jk) )           &
                        &                            * ( ptb(ji,jj,jk-1,jn) - ptb(ji,jj,jk,jn) )
                  END DO
               END DO
            END DO
            !
         ELSE                                   ! bilaplacian 
            SELECT CASE( kpass )
            CASE(  1  )                            ! 1st pass : eddy coef = zah_wslp2
               DO jk = tile % jsk_2, tile % jekp1 
                  DO jj = tile % jsj, tile % jej
                     DO ji = tile % jsi, tile % jei
                        ztfw(ji,jj,jk) = ztfw(ji,jj,jk)                       &
                           &           + zah_wslp2(ji,jj,jk) * e1e2t(ji,jj)   &
                           &           * ( ptb(ji,jj,jk-1,jn) - ptb(ji,jj,jk,jn) ) / e3w_n(ji,jj,jk) * wmask(ji,jj,jk)
                     END DO
                  END DO
               END DO 
            CASE(  2  )                         ! 2nd pass : eddy flux = zah_wslp2 and akz applied on ptb  and ptbb gradients, resp.
               DO jk = tile % jsk_2, tile % jekp1
                  DO jj = tile % jsj, tile % jej
                     DO ji = tile % jsi, tile % jei
                        ztfw(ji,jj,jk) = ztfw(ji,jj,jk) + e1e2t(ji,jj) / e3w_n(ji,jj,jk) * wmask(ji,jj,jk)                       &
                           &                            * (  zah_wslp2(ji,jj,jk) * ( ptb (ji,jj,jk-1,jn) - ptb (ji,jj,jk,jn) )   &
                           &                               +      zakz(ji,jj,jk) * ( ptbb(ji,jj,jk-1,jn) - ptbb(ji,jj,jk,jn) )   )
                     END DO
                  END DO
               END DO
            END SELECT
         ENDIF
         !         
         DO jk = tile % jsk, tile % jek                 !==  Divergence of vertical fluxes added to pta  ==!
            DO jj = tile % jsj, tile % jej
               DO ji = tile % jsi, tile % jei   ! vector opt.
                  pta(ji,jj,jk,jn) = pta(ji,jj,jk,jn) + zsign * (  ztfw (ji,jj,jk) - ztfw(ji,jj,jk+1)  )   &
                     &                                        * r1_e1e2t(ji,jj) / e3t_n(ji,jj,jk)
               END DO
            END DO
         END DO
         !
         IF ( l_ptr .OR. l_hst ) THEN 
            DO jk = tile % jsk, tile % jek                 
               DO jj = tile % jsj, tile % jej
                  DO ji = tile % jsi, tile % jei   ! vector opt.
                     pftu(ji,jj,jk,jn) = zftu(ji,jj,jk) ; pftv(ji,jj,jk,jn) = zftv(ji,jj,jk)
                  END DO
               END DO
            END DO
         END IF 
 
         !                                                        ! ===============
      END DO                                                      ! end tracer loop
      !
!$ACC END KERNELS
   END SUBROUTINE tra_ldf_iso_tile

   !!==============================================================================
END MODULE traldf_iso_tile
